--[[
	ManualDischarge.lua
	
	Autor: 		Ifko[nator]
	Datum: 		01.04.2020
	Version: 	1.0
	
	Changelog:	V 1.0 @01.04.2020 - initial implementation in FS 19
]]

ManualDischarge = {};

function ManualDischarge.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Dischargeable, specializations);
end;

function ManualDischarge.registerEventListeners(vehicleType)
	local functionNames = {
		"onPreLoad",
		"onLoad",
		"onDraw",
		"onRegisterActionEvents",
		"onWriteStream",
		"onReadStream",
		"onUpdateTick"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, ManualDischarge);
	end;
end;

function ManualDischarge.registerFunctions(vehicleType)
	local newFunctions = {
		"setAllowDischarge",
		"foundObjectToDischarge",
		"objectToDischargeHasFreeCapacity"
	};
	
	for _, newFunction in ipairs(newFunctions) do
		SpecializationUtil.registerFunction(vehicleType, newFunction, ManualDischarge[newFunction]);
	end;
end;

function ManualDischarge.registerOverwrittenFunctions(vehicleType)
    local overwrittenFunctions = {
		"getCanDischargeToObject",
		"getDoConsumePtoPower",
		"getIsPowerTakeOffActive"
	};
	
	for _, overwrittenFunction in ipairs(overwrittenFunctions) do
		SpecializationUtil.registerOverwrittenFunction(vehicleType, overwrittenFunction, ManualDischarge[overwrittenFunction]);
	end;
end;

function ManualDischarge:onPreLoad(savegame)
	local specManualDischarge = self.spec_manualDischarge;
	local configurationId = self.configurations["manualDischarge"];
	
    if configurationId ~= nil then
		local storeItem = g_storeManager:getItemByXMLFilename(self.configFileName);
		
		if storeItem ~= nil then
            local configuration = storeItem.configurations["manualDischarge"][configurationId];

            if configuration ~= nil then
                specManualDischarge.isBought = configuration.index == 2;
            end;
        end;
    end;
end;

function ManualDischarge:onLoad(savegame)
	local specManualDischarge = self.spec_manualDischarge;
	local specFillUnit = self.spec_fillUnit;
	
	specManualDischarge.allowDischarge = false;
	specManualDischarge.showTurnOnMotorWarning = false;
	specManualDischarge.showVehicleIsEmptyWarning = false;

	if specManualDischarge.isBought then
		if Pipe ~= nil then	
			Pipe.onDischargeStateChanged = Utils.overwrittenFunction(Pipe.onDischargeStateChanged, ManualDischarge.onDischargeStateChanged);
		end;
	else
		specManualDischarge.allowDischarge = true;
	end;
	
	self:setAllowDischarge(specManualDischarge.allowDischarge);
end;

function ManualDischarge:getDoConsumePtoPower(superFunc)
	local specManualDischarge = self.spec_manualDischarge;

	if not specManualDischarge.isBought then
		specManualDischarge.allowDischarge = true;
	end;

    return specManualDischarge.allowDischarge and self:getDischargeState() ~= Dischargeable.DISCHARGE_STATE_OFF or self.getIsTurnedOn ~= nil and self:getIsTurnedOn();
end;

function ManualDischarge:getIsPowerTakeOffActive(superFunc)
	local specManualDischarge = self.spec_manualDischarge;

	if not specManualDischarge.isBought then
		specManualDischarge.allowDischarge = true;
	end;

    return specManualDischarge.allowDischarge and self:getDischargeState() ~= Dischargeable.DISCHARGE_STATE_OFF or self.getIsTurnedOn ~= nil and self:getIsTurnedOn();
end;

function ManualDischarge:onDischargeStateChanged(superFunc, state)
    if self.isClient then
		local specPipe = self.spec_pipe;
		local specManualDischarge = self.spec_manualDischarge;

		if not specManualDischarge.isBought then
			specManualDischarge.allowDischarge = true;
		end;

		if specManualDischarge ~= nil then
        	local dischargeNode = self:getCurrentDischargeNode();
			local dischargeNodeIndex = nil;

        	if dischargeNode ~= nil then
        	    dischargeNodeIndex = dischargeNode.index;
			end;

        	if dischargeNodeIndex == specPipe.dischargeNodeIndex then
        	    if state == Dischargeable.DISCHARGE_STATE_OFF or not specManualDischarge.allowDischarge then
        	        g_animationManager:stopAnimations(specPipe.animationNodes);
        	    else
        	        g_animationManager:startAnimations(specPipe.animationNodes);
        	        g_animationManager:setFillType(specPipe.animationNodes, self:getFillUnitLastValidFillType(dischargeNode.fillUnitIndex));
        	    end;
			end;
		else
			superFunc(self, state);
		end;
    end;
end;

function ManualDischarge:onUpdateTick(dt, isActiveForInput, isActiveForInputIngnoreSelection, isSelected)
	local specManualDischarge = self.spec_manualDischarge;
	
	if specManualDischarge.isBought then
		local specDischargeable = self.spec_dischargeable;
		local specCombine = self.spec_combine;
		local specTrailer = self.spec_trailer;
		local specPipe = self.spec_pipe;
		
		local currentDischargeNode = specDischargeable.currentDischargeNode;

		if currentDischargeNode ~= nil then
			local allowTipToGround = self:getCanToggleDischargeToGround() and self:getCanDischargeToGround(currentDischargeNode) and self:getCanDischargeAtPosition(currentDischargeNode);
			local allowDischarge = true;

			if currentDischargeNode.dischargeObject ~= nil then
				allowDischarge = self:objectToDischargeHasFreeCapacity(currentDischargeNode);
			end;

			local pipeIsUnFolded = specPipe.targetState > 1;

			if specPipe.animation.name ~= nil then
				pipeIsUnFolded = pipeIsUnFolded and not self:getIsAnimationPlaying(specPipe.animation.name);
			end;

			if specCombine ~= nil then	
				if self:getIsAIActive() then
					self:setAllowDischarge(self:getFillUnitFillLevel(specCombine.fillUnitIndex) > 0 and not allowTipToGround and pipeIsUnFolded);
				end;
			
				if self.getIsCourseplayDriving ~= nil then
					if self:getIsCourseplayDriving() then
						self:setAllowDischarge(self:getFillUnitFillLevel(specCombine.fillUnitIndex) > 0 and not allowTipToGround and pipeIsUnFolded);
					end;
				end;
			end;

			if self.getAttacherVehicle ~= nil then
				local controlledTractor = self:getAttacherVehicle();

				if controlledTractor ~= nil then
					if controlledTractor.getIsCourseplayDriving ~= nil then
						if controlledTractor:getIsCourseplayDriving() then
							if specTrailer ~= nil and specPipe ~= nil then	
								self:setAllowDischarge(self:getFillUnitFillLevel(currentDischargeNode.fillUnitIndex) < 0.0001 and not allowTipToGround and pipeIsUnFolded);
							end;
						end;
					end;
				end;
			end;

			if not currentDischargeNode.stopDischargeOnEmpty or hasXMLProperty(self.xmlFile, "vehicle.woodCrusher") and not specManualDischarge.allowDischarge then
				--## forage harevester or wood chrusher, no button needed to overload
				
				self:setAllowDischarge(true);
			end;

			specManualDischarge.showTurnOnMotorWarning = false;
			specManualDischarge.showVehicleIsEmptyWarning = false;

			if specManualDischarge.allowDischarge then
				if currentDischargeNode.stopDischargeOnEmpty then 
					if specCombine ~= nil then
						if self:getFillUnitFillLevel(specCombine.fillUnitIndex) < 0.0001 then
							allowDischarge = false;
							allowTipToGround = false;

							specManualDischarge.showVehicleIsEmptyWarning = true;
						end;

						if self.getIsMotorStarted ~= nil then
							if not self:getIsMotorStarted() then
								allowDischarge = false;
								allowTipToGround = false;

								specManualDischarge.showTurnOnMotorWarning = true;
							end;
						end;
					end;

					if specTrailer ~= nil and specPipe ~= nil then
						if self:getFillUnitFillLevel(currentDischargeNode.fillUnitIndex) < 0.0001 then
							allowDischarge = false;
							allowTipToGround = false;

							specManualDischarge.showVehicleIsEmptyWarning = true;
						end;
					end;

					if self.getAttacherVehicle ~= nil then
						local controlledTractor = self:getAttacherVehicle();

						if controlledTractor ~= nil then
							if not controlledTractor:getIsMotorStarted() then
								allowDischarge = false;
								allowTipToGround = false;

								specManualDischarge.showTurnOnMotorWarning = true;
							end;
						end;
					end;

					if allowTipToGround then
						self:setDischargeState(Dischargeable.DISCHARGE_STATE_GROUND);
					else
						self:setAllowDischarge(allowDischarge);
					end;
				end;
			end;

			if currentDischargeNode.stopDischargeOnEmpty then
    			local allowShowButton = allowDischarge and currentDischargeNode.stopDischargeOnEmpty and self:getCanDischargeAtPosition(currentDischargeNode);
				local dischargeButton = specManualDischarge.actionEvents[InputAction.DISCHARGE_BUTTON];
				local actionEventTip = specDischargeable.actionEvents[InputAction.TOGGLE_TIPSTATE];
				local actionEventTipGround = specDischargeable.actionEvents[InputAction.TOGGLE_TIPSTATE_GROUND];
				
				if specPipe ~= nil then
					allowShowButton = allowShowButton and pipeIsUnFolded;
				end;

				if actionEventTip ~= nil then
					g_inputBinding:setActionEventTextVisibility(actionEventTip.actionEventId, false);
					g_inputBinding:setActionEventActive(actionEventTip.actionEventId, false);
				end;

    			if actionEventTipGround ~= nil then
					g_inputBinding:setActionEventTextVisibility(actionEventTipGround.actionEventId, false);
					g_inputBinding:setActionEventActive(actionEventTipGround.actionEventId, false);
    			end;

				if dischargeButton ~= nil then
					local currentText = Vehicle.START_DISCHARGE;

					g_inputBinding:setActionEventActive(dischargeButton.actionEventId, allowShowButton);
					g_inputBinding:setActionEventTextVisibility(dischargeButton.actionEventId, allowShowButton);

					if specManualDischarge.allowDischarge then
						currentText = Vehicle.STOP_DISCHARGE;
					else
						self:setDischargeEffectActive(currentDischargeNode, false, true);
						self:updateDischargeSound(currentDischargeNode, 99999);
					end;

					g_inputBinding:setActionEventText(dischargeButton.actionEventId, currentText);
				end;
			end;
		end;
	end;
end;

function ManualDischarge:onDraw(isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local specManualDischarge = self.spec_manualDischarge;
	
	if specManualDischarge.isBought then
		local specTrailer = self.spec_trailer;
		local specPipe = self.spec_pipe;
		
    	if specManualDischarge.showTurnOnMotorWarning then
    	    g_currentMission:showBlinkingWarning(g_i18n:getText("warning_motorNotStarted"), 2000);
		end;

		if specManualDischarge.showVehicleIsEmptyWarning then
			local warningText = Vehicle.COMBINE_EMPTY_WARNING;

			if specTrailer ~= nil and specPipe ~= nil then
				warningText = Vehicle.AUGERWAGON_EMPTY_WARNING;
			end;

			g_currentMission:showBlinkingWarning(warningText, 2000);
		end;
	end;
end;

function ManualDischarge:objectToDischargeHasFreeCapacity(dischargeNode)
	local object = dischargeNode.dischargeObject;
	
    if object == nil then
        return false;
	end;
	
	local fillType = self:getDischargeFillType(dischargeNode);
	
    if not object:getFillUnitSupportsFillType(dischargeNode.dischargeFillUnitIndex, fillType) then
        return false;
	end;
	
    if not object:getFillUnitAllowsFillType(dischargeNode.dischargeFillUnitIndex, fillType) then
        return false;
	end;
	
    if object.getFillUnitFreeCapacity ~= nil and object:getFillUnitFreeCapacity(dischargeNode.dischargeFillUnitIndex, fillType, self:getActiveFarm()) <= 0 then
        return false;
	end;
	
    if object.getIsFillAllowedFromFarm ~= nil and not object:getIsFillAllowedFromFarm(self:getActiveFarm()) then
        return false;
	end;

	return true;
end;

function ManualDischarge:foundObjectToDischarge(dischargeNode)
	local specManualDischarge = self.spec_manualDischarge;

	if dischargeNode == nil then
        return false;
	end;
	
	if not self:objectToDischargeHasFreeCapacity(dischargeNode) then
		return false;
	end;

    --## Adding should only be done if removing is allowed (generally adding is always allowed because it is beneficial to the receiver)
    --## A case where this is not allowed is when this is a pallet where the the controller does not own it (or can access it)
    if self.getMountObject ~= nil then
		local mounter = self:getDynamicMountObject() or self:getMountObject();
		
        if mounter ~= nil then
            --## if the active farm of the mounter has NO access to farmId fill unit: disallow
            if not g_currentMission.accessHandler:canFarmAccess(mounter:getActiveFarm(), self, true) then
                return false;
            end;
        end;
	end;
	
	return true;
end;

function ManualDischarge:getCanDischargeToObject(superFunc, dischargeNode)
	local specManualDischarge = self.spec_manualDischarge;

	if not specManualDischarge.isBought then
		specManualDischarge.allowDischarge = true;
	end;

	if self.getIsEntered ~= nil then
		if not specManualDischarge.allowDischarge and not self:getIsEntered() then
			return false;
		end;
	end;

	return self:foundObjectToDischarge(dischargeNode) and specManualDischarge.allowDischarge;
end;

function ManualDischarge:setAllowDischarge(allowDischarge, noEventSend)
	local specManualDischarge = self.spec_manualDischarge;
	local specPipe = self.spec_pipe;

    if allowDischarge ~= specManualDischarge.allowDischarge then
		if not noEventSend then
			if g_server ~= nil then
				g_server:broadcastEvent(ManualDischargeEvent:new(self, allowDischarge), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(ManualDischargeEvent:new(self, allowDischarge));
			end;
		end;

		if allowDischarge then
			self:setDischargeState(Dischargeable.DISCHARGE_STATE_OBJECT);

			if specPipe ~= nil then
				g_animationManager:startAnimations(specPipe.animationNodes);
			end;
		else
			self:setDischargeState(Dischargeable.DISCHARGE_STATE_OFF);
		end;

		specManualDischarge.allowDischarge = allowDischarge;
	end;
end;

function ManualDischarge:onRegisterActionEvents(isActiveForInput)
	if self.isClient then
        local specManualDischarge = self.spec_manualDischarge;
		
		if specManualDischarge.isBought then
			self:clearActionEventsTable(specManualDischarge.actionEvents);
			
			if self:getIsActiveForInput(true) then
				local _, actionEventId = self:addActionEvent(specManualDischarge.actionEvents, InputAction.DISCHARGE_BUTTON, self, ManualDischarge.actionEventDischarge, false, true, false, true, nil);

				g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
				g_inputBinding:setActionEventTextVisibility(actionEventId, true);
				g_inputBinding:setActionEventText(actionEventId, Vehicle.START_DISCHARGE);
				g_inputBinding:setActionEventActive(actionEventId, false);
			end;
		end;
	end;
end;

function ManualDischarge.actionEventDischarge(self, actionName, inputValue, callbackState, isAnalog)
	local specManualDischarge = self.spec_manualDischarge;
		
	self:setAllowDischarge(not specManualDischarge.allowDischarge);
end;

function ManualDischarge:onWriteStream(streamId, connection)
    if not connection:getIsServer() then 
		local specManualDischarge = self.spec_manualDischarge;

		if not specManualDischarge.isBought then
			specManualDischarge.allowDischarge = true;
		end;
		
		streamWriteBool(streamId, specManualDischarge.allowDischarge);
	end;
end;

function ManualDischarge:onReadStream(streamId, connection)
    if connection:getIsServer() then
		local specManualDischarge = self.spec_manualDischarge;

		if not specManualDischarge.isBought then
			specManualDischarge.allowDischarge = true;
		end;
		
		specManualDischarge.allowDischarge = streamReadBool(streamId);
	end;
end;

ManualDischargeEvent = {};
ManualDischargeEvent_mt = Class(ManualDischargeEvent, Event);

InitEventClass(ManualDischargeEvent, "ManualDischargeEvent");

function ManualDischargeEvent:emptyNew()
	local self = Event:new(ManualDischargeEvent_mt);
    
	return self;
end;

function ManualDischargeEvent:new(vehicle, allowDischarge)
	local self = ManualDischargeEvent:emptyNew();
	
	self.vehicle = vehicle;
	self.allowDischarge = allowDischarge;

	return self;
end;

function ManualDischargeEvent:readStream(streamId, connection)
	self.vehicle = NetworkUtil.readNodeObject(streamId);
	self.allowDischarge = streamReadBool(streamId);
	
	self:run(connection);
end;

function ManualDischargeEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.vehicle);

	streamWriteBool(streamId, self.allowDischarge);
end;

function ManualDischargeEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(ManualDischargeEvent:new(self.vehicle, self.allowDischarge), nil, connection, self.vehicle);
	end;
	
    if self.vehicle ~= nil then
        self.vehicle:setAllowDischarge(self.allowDischarge, true);
	end;
end;